/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef EMUCONTROLLER_H
#define EMUCONTROLLER_H

#import <Cocoa/Cocoa.h>
#include "CrabEmu.h"
#include "PaletteGLView.h"
#include "PatternGLView.h"
#include "joystick.h"

@interface EmuController : NSObject
{
    IBOutlet NSMenuItem *noise;
    IBOutlet NSMenuItem *pause;
    IBOutlet NSMenuItem *reset;
    IBOutlet NSMenuItem *PSG;
    IBOutlet NSMenuItem *tone0;
    IBOutlet NSMenuItem *tone1;
    IBOutlet NSMenuItem *tone2;
    IBOutlet NSMenuItem *debugger;
    IBOutlet NSPanel *virtKeyboard;
    IBOutlet NSPanel *prefWindow;
    IBOutlet NSWindow *debuggerWindow;
    IBOutlet NSWindow *smsVdpDebuggerWindow;
    IBOutlet NSMenuItem *rExport;
    IBOutlet NSMenuItem *rDomestic;
    IBOutlet NSMenuItem *pal;
    IBOutlet NSMenuItem *ntsc;
    IBOutlet NSMenu *p1input;
    IBOutlet NSMenu *p2input;
    IBOutlet NSMenuItem *p1keyboard;
    IBOutlet NSMenuItem *p2keyboard;

    IBOutlet NSPopUpButton *pjoysticks;
    IBOutlet NSButton *pjoyUp;
    IBOutlet NSButton *pjoyRight;
    IBOutlet NSButton *pjoyDown;
    IBOutlet NSButton *pjoyLeft;
    IBOutlet NSButton *pjoyB1;
    IBOutlet NSButton *pjoyB2;
    IBOutlet NSButton *pjoyGGStart;
    IBOutlet NSPanel *joyPanel;

    IBOutlet NSTextField *smsBios;
    IBOutlet NSButton *smsBiosEnabled;
    IBOutlet NSTextField *ggBios;
    IBOutlet NSButton *ggBiosEnabled;
    IBOutlet NSButton *ssaverDisabled;

    IBOutlet NSMenu *saveStates;
    IBOutlet NSMenu *loadStates;
    IBOutlet NSMenuItem *saveSRAM;
    IBOutlet NSMenuItem *screenshot;
    IBOutlet NSMenuItem *fullscreen;
    IBOutlet NSMenuItem *showVDPPalette;
    IBOutlet NSMenuItem *showVDPPatterns;

    IBOutlet NSPanel *vdpPalettePanel;
    IBOutlet PaletteGLView *paletteView;

    IBOutlet NSPanel *vdpPatternPanel;
    IBOutlet PatternGLView *patternView;

    NSMenuItem *p1curinput;
    NSMenuItem *p2curinput;

    NSUserDefaults *prefs;
    NSString *buttonmap;
    NSString *openRom;
    NSMutableDictionary *keymap;

    int joy_count;

    joydata_t *joy1;
    NSMutableDictionary *j1map;

    joydata_t *joy2;
    NSMutableDictionary *j2map;

    NSTimer *timer;

    BOOL domestic;
}

/* NSApplication delegate methods */
- (void)application:(NSApplication *)app openFiles:(NSArray *)filenames;
- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app;

- (void)awakeFromNib;
- (void)dealloc;

- (IBAction)enableNoise:(id)sender;
- (IBAction)enablePSG:(id)sender;
- (IBAction)enableTone0:(id)sender;
- (IBAction)enableTone1:(id)sender;
- (IBAction)enableTone2:(id)sender;
- (IBAction)loadMasterRom:(id)sender;
- (IBAction)prefGGStart:(id)sender;
- (IBAction)prefSms1B1:(id)sender;
- (IBAction)prefSms1B2:(id)sender;
- (IBAction)prefSms1Down:(id)sender;
- (IBAction)prefSms1Left:(id)sender;
- (IBAction)prefSms1Reset:(id)sender;
- (IBAction)prefSms1Right:(id)sender;
- (IBAction)prefSms1Up:(id)sender;
- (IBAction)prefSms2B1:(id)sender;
- (IBAction)prefSms2B2:(id)sender;
- (IBAction)prefSms2Down:(id)sender;
- (IBAction)prefSms2Left:(id)sender;
- (IBAction)prefSms2Reset:(id)sender;
- (IBAction)prefSms2Right:(id)sender;
- (IBAction)prefSms2Up:(id)sender;
- (IBAction)enableDebugger:(id)sender;
- (IBAction)saveSRAM:(id)sender;
- (IBAction)saveState1:(id)sender;
- (IBAction)saveState2:(id)sender;
- (IBAction)saveState3:(id)sender;
- (IBAction)saveState4:(id)sender;
- (IBAction)saveState5:(id)sender;
- (IBAction)saveState6:(id)sender;
- (IBAction)loadState1:(id)sender;
- (IBAction)loadState2:(id)sender;
- (IBAction)loadState3:(id)sender;
- (IBAction)loadState4:(id)sender;
- (IBAction)loadState5:(id)sender;
- (IBAction)loadState6:(id)sender;
- (IBAction)resetEmulation:(id)sender;
- (IBAction)regionDomestic:(id)sender;
- (IBAction)regionExport:(id)sender;
- (IBAction)modeNTSC:(id)sender;
- (IBAction)modePAL:(id)sender;
- (IBAction)scanJoysticks:(id)sender;
- (IBAction)setInput1:(id)sender;
- (IBAction)setInput2:(id)sender;
- (IBAction)prefJoyButton:(id)sender;
- (IBAction)prefSMSBiosBrowse:(id)sender;
- (IBAction)prefGGBiosBrowse:(id)sender;
- (IBAction)prefBiosEnabledClicked:(id)sender;
- (IBAction)takeScreenshot:(id)sender;
- (IBAction)toggleFullscreen:(id)sender;
- (IBAction)toggleScreensaver:(id)sender;
- (IBAction)showVDPPalette:(id)sender;
- (IBAction)showVDPPatterns:(id)sender;

- (void)handleSMSButton:(int)button withJoystickElement:(NSNumber *)n;
- (void)handleJoysticks;
- (void)updatePalette;
- (void)updatePatterns;

- (unichar)runVirtKeyboard;
- (void)endVirtKeyboard:(unichar)character;

- (void)setJoystick:(joydata_t *)joy
             button:(int)button
         assignment:(uint32)a;
- (void)joyInputThread:(id)b;

- (void)masterEmuThread:(id)junk;
- (void)pauseEmulation:(id)sender;

- (NSString *)buttonMap;

- (BOOL)initMasterSystem;
- (BOOL)loadMasterSystemRomFromFile:(NSString *)filename;

- (void)fillDictionary:(NSMutableDictionary *)d withButtons:(unichar *)string;
- (void)fillString:(unichar *)string withDictionary:(NSDictionary *)d;

- (void)savePrefs;

- (void)ssTimer:(NSTimer *)timer;

@end

#endif
