/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "EmuOGLView.h"
#include "EmuController.h"

#include "sms.h"
#include "smsvdp.h"
#include "terebi.h"

extern EmuOGLView *theview;
extern void render_scene();
extern void gui_resized(void);

@implementation EmuOGLView

- (id)initWithFrame:(NSRect)frameRect
{
    CGLContextObj cxt = CGLGetCurrentContext();

    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    if(CGLLockContext != NULL)  {
        CGLLockContext(cxt);
    }

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs];

    if(fmt == nil)  {
        if(CGLUnlockContext != NULL)    {
            CGLUnlockContext(cxt);
        }
        return nil;
    }

    [fmt autorelease];

    if((self = [super initWithFrame:frameRect pixelFormat:fmt]) != nil) {
        theview = self;
    }
    else    {
        if(CGLUnlockContext != NULL)    {
            CGLUnlockContext(cxt);
        }
        return nil;
    }

    /* Set up the full screen context. */
    disp = [self screenID];
    [self createFullscreenContext];

    resized = NO;
    isFullscreen = NO;

    if(CGLUnlockContext != NULL)    {
        CGLUnlockContext(cxt);
    }
    return self;
}

- (void)drawScene
{
    CGLContextObj cxt = CGLGetCurrentContext();
    static BOOL wasFullscreen = NO;
    static int count = 2;

    if(CGLLockContext != NULL)  {
        CGLLockContext(cxt);
    }

    if(resized) {
        glViewport(0.0f, 0.0f, [self width], [self height]);
    }

    if((!wasFullscreen && isFullscreen) ||
       (wasFullscreen && !isFullscreen))    {
        gui_resized();

        /* UGLY UGLY UGLY HACK. For some reason, fullscreen mode seems to need
           some extra encouragement when switching into it sometimes. Make sure
           that gui_resized() gets called at least a few times with the new
           sizes set up. */
        if(!(--count))  {
            wasFullscreen = isFullscreen;
            count = 2;
        }
    }

    render_scene();

    if(!isFullscreen)   {
        [[self openGLContext] flushBuffer];
    }
    else    {
        [fsCxt flushBuffer];
    }

    if(CGLUnlockContext != NULL)    {
        CGLUnlockContext(cxt);
    }

    [controller handleJoysticks];

    if(!isFullscreen || [[NSScreen screens] count] > 1) {
        [controller updatePalette];
        [controller updatePatterns];
    }
}

- (void)drawRect:(NSRect)rect
{
    [self drawScene];
}

- (void)setNeedsDisplay:(BOOL)display
{
    if(isFullscreen)    {
        if(display) {
            [fsCxt makeCurrentContext];
            [self drawScene];
        }
    }
    else    {
        [super setNeedsDisplay:display];
    }
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- (void)showWindow
{
    [window makeKeyAndOrderFront:self];
    [window setContentAspectRatio:NSMakeSize(4.0, 3.0)];
}

- (void)keyDown:(NSEvent *)event
{
    unichar buttonmap[14];
    unichar character = (unichar)[event keyCode];

    [[controller buttonMap] getCharacters:buttonmap];

    if(character == buttonmap[0])
        sms_button_pressed(SMS_PAD1_UP);
    else if(character == buttonmap[1])
        sms_button_pressed(SMS_PAD1_DOWN);
    else if(character == buttonmap[2])
        sms_button_pressed(SMS_PAD1_LEFT);
    else if(character == buttonmap[3])
        sms_button_pressed(SMS_PAD1_RIGHT);
    else if(character == buttonmap[4])
        sms_button_pressed(SMS_PAD1_A);
    else if(character == buttonmap[5])
        sms_button_pressed(SMS_PAD1_B);
    else if(character == buttonmap[6])
        sms_button_pressed(GG_START);
    else if(character == buttonmap[7])
        sms_button_pressed(SMS_PAD2_UP);
    else if(character == buttonmap[8])
        sms_button_pressed(SMS_PAD2_DOWN);
    else if(character == buttonmap[9])
        sms_button_pressed(SMS_PAD2_LEFT);
    else if(character == buttonmap[10])
        sms_button_pressed(SMS_PAD2_RIGHT);
    else if(character == buttonmap[11])
        sms_button_pressed(SMS_PAD2_A);
    else if(character == buttonmap[12])
        sms_button_pressed(SMS_PAD2_B);
    else
        [super keyDown: event];
}

- (void)keyUp:(NSEvent *)event
{
    unichar buttonmap[14];
    unichar character = (unichar)[event keyCode];

    [[controller buttonMap] getCharacters:buttonmap];

    if(character == buttonmap[0])
        sms_button_released(SMS_PAD1_UP);
    else if(character == buttonmap[1])
        sms_button_released(SMS_PAD1_DOWN);
    else if(character == buttonmap[2])
        sms_button_released(SMS_PAD1_LEFT);
    else if(character == buttonmap[3])
        sms_button_released(SMS_PAD1_RIGHT);
    else if(character == buttonmap[4])
        sms_button_released(SMS_PAD1_A);
    else if(character == buttonmap[5])
        sms_button_released(SMS_PAD1_B);
    else if(character == buttonmap[6])
        sms_button_released(GG_START);
    else if(character == buttonmap[7])
        sms_button_released(SMS_PAD2_UP);
    else if(character == buttonmap[8])
        sms_button_released(SMS_PAD2_DOWN);
    else if(character == buttonmap[9])
        sms_button_released(SMS_PAD2_LEFT);
    else if(character == buttonmap[10])
        sms_button_released(SMS_PAD2_RIGHT);
    else if(character == buttonmap[11])
        sms_button_released(SMS_PAD2_A);
    else if(character == buttonmap[12])
        sms_button_released(SMS_PAD2_B);
    else
        [super keyUp: event];
}

- (void)mouseDown:(NSEvent *)event
{
    NSPoint loc = [self convertPoint:[event locationInWindow]
                            fromView:nil];
    NSRect bounds = [self bounds];
    int x, y;

    /* Scale the position to [0.0, 1.0], and multiply by the SG-1000's
       resolution. */
    x = (int)((loc.x / bounds.size.width) * 256);
    y = (int)((loc.y / bounds.size.height) * 192);

    /* Mac OS X takes positions from the lower-left corner, we want it from the
       upper left. Flip the y value so that we get that. */
    y = (192 - y) > 0 ? 192 - y : 0;

    terebi_update(x, y, 1);
}

- (void)mouseUp:(NSEvent *)event
{
    terebi_update(0, 0, 0);
}

- (void)mouseDragged:(NSEvent *)event
{
    [self mouseDown:event];
}

- (float)width
{
    if(!isFullscreen)   {
        return [self bounds].size.width;
    }
    else    {
        return [[NSScreen mainScreen] frame].size.width;
    }
}

- (float)height
{
    if(!isFullscreen)   {
        return [self bounds].size.height;
    }
    else    {
        return [[NSScreen mainScreen] frame].size.height;
    }
}

- (void)setAspectRatioX:(float)x Y:(float)y
{
    /* FIXME: Resize window to be in the correct aspect ratio. */
    [window setContentAspectRatio:NSMakeSize(x, y)];
}

- (void)update
{
    /* Set a flag here, and actually handle the resizing in drawRect, since
       the bounds have not changed yet when update is called. */
    resized = YES;

    [super update];
}

- (void)setTitle:(NSString *)str
{
    [window setTitle:str];
}

- (BOOL)isFullscreen
{
    return isFullscreen;
}

- (void)toggleFullscreen
{
    CGError err;
    CGDisplayFadeReservationToken token;

    /* This probably will crash on Mac OS X 10.3, so make sure not to crash by
       not allowing fullscreen mode there. */
    if(CGLLockContext == NULL || CGLUnlockContext == NULL)
        return;

    err = CGAcquireDisplayFadeReservation(kCGMaxDisplayReservationInterval,
                                          &token);

    if(err == kCGErrorSuccess)  {
        CGDisplayFade(token, 0.5, kCGDisplayBlendNormal,
                      kCGDisplayBlendSolidColor, 0, 0, 0, 1);
    }

    if(isFullscreen)    {
        [fsCxt clearDrawable];
        [[self openGLContext] makeCurrentContext];
        CGReleaseAllDisplays();
    }
    else    {
        CGDirectDisplayID d = [self screenID];

        if(d != disp)   {
            [fsCxt release];
            disp = d;
            [self createFullscreenContext];
        }

        [fsCxt clearDrawable];
        CGDisplayCapture(disp);
        [fsCxt setFullScreen];
        [fsCxt makeCurrentContext];
    }

    if(err == kCGErrorSuccess)  {
        CGDisplayFade(token, 0.5, kCGDisplayBlendNormal,
                      kCGDisplayBlendSolidColor, 0, 0, 0, 0);
        CGReleaseDisplayFadeReservation(token);
    }

    isFullscreen = !isFullscreen;
}

- (CGDirectDisplayID)screenID
{
    NSArray *screens = [NSScreen screens];
    NSEnumerator *i = [screens objectEnumerator];
    NSScreen *obj;
    NSRect f = [window frame];
    NSRect sf;

    while((obj = (NSScreen *)[i nextObject]))   {
        sf = [obj frame];

        if(f.origin.x >= sf.origin.x && f.origin.y >= sf.origin.y &&
           f.origin.x <= sf.origin.x + sf.size.width &&
           f.origin.y <= sf.origin.y + sf.size.height)  {
            NSDictionary *d = [obj deviceDescription];
            NSNumber *n = (NSNumber *)[d objectForKey:@"NSScreenNumber"];

            return (CGDirectDisplayID)[n unsignedIntValue];
        }
    }

    /* Punt. */
    return kCGDirectMainDisplay;
}

- (void)createFullscreenContext
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAFullScreen,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFAScreenMask,
        CGDisplayIDToOpenGLDisplayMask(disp),
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    if(fmt != nil)  {
        long sync = 1;

        fsCxt = [[NSOpenGLContext alloc] initWithFormat:fmt
                                           shareContext:[self openGLContext]];
        [fmt release];
        [fsCxt setValues:&sync forParameter:NSOpenGLCPSwapInterval];
    }
    else {
        fsCxt = nil;
    }
}

@end
