/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <OpenGL/GL.h>
#include "PaletteGLView.h"

#include "smsvdp.h"

/* From sms.c */
extern int sms_initialized;

@implementation PaletteGLView

- (id)initWithFrame:(NSRect)frameRect
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes: attrs];
    
    if(fmt == nil)
        return nil;

    [fmt autorelease];

    if((self = [super initWithFrame:frameRect pixelFormat:fmt]) == nil) {
        return nil;
    }

    return self;
}

- (void)drawRect:(NSRect)rect
{
    static BOOL first = YES;
    int i;
    sms_vdp_color_t col;

    if(first)   {
        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(1.0f);

        glViewport(0.0f, 0.0f, [self bounds].size.width,
                   [self bounds].size.height);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        glOrtho(0.0f, 320.0f, 120.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        first = NO;
    }

    if(sms_initialized) {
        for(i = 0; i < 32; ++i) {
            col = smsvdp.pal[i];

            glBegin(GL_QUADS);
            glColor3ub((col >> 16) & 0xFF, (col >> 8) & 0xFF, col & 0xFF);
            glVertex2f(i * 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 120.0f);
            glVertex2f(i * 10.0f, 120.0f);
            glEnd();
        }
    }

    glFlush();
    [[self openGLContext] flushBuffer];
}

@end
