/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "CrabEmuThread.h"

extern int emu_frame(int run);
extern void render_scene(void);
extern double GetAbsoluteTime(void);
extern int sms_initialized;

CrabEmuThread::CrabEmuThread(QObject *parent) : QThread(parent) {
    finished = false;
}

CrabEmuThread::~CrabEmuThread() {
    finish();
    wait();
}

void CrabEmuThread::run()   {
    double timer = GetAbsoluteTime();
    int run = 0;

    while(!finished)    {
        if(!sms_initialized)    {
            sleep(1);
            continue;
        }

        run = emu_frame(run);
        render_scene();
        emit frameRendered();
        
        while(timer + 1.0 / 60.0 > GetAbsoluteTime())
            usleep(10);
        
        timer = GetAbsoluteTime();
    }
}

void CrabEmuThread::finish()    {
    finished = 1;
}
