/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QPainter>
#include <QMouseEvent>

#include "EmuWidget.h"
#include "CrabEmuThread.h"
#include "CrabEmu.h"

#include "sms.h"
#include "smsvdp.h"
#include "terebi.h"

extern int sms_console;
extern sms_vdp_t smsvdp;

EmuWidget::EmuWidget(CrabEmuThread *thd, QWidget *parent) : QWidget(parent),
img(256, 256, QImage::Format_RGB32) {
    connect(thd, SIGNAL(frameRendered()), this, SLOT(update()));
}

QImage *EmuWidget::retainImage()    {
    mutex.lock();
    return &img;
}

void EmuWidget::releaseImage()  {
    mutex.unlock();
}

void EmuWidget::paintEvent(QPaintEvent *event __UNUSED__)   {
    QMutexLocker locker(&mutex);
    QPainter painter(this);
    QRect r;

    if(sms_console == CONSOLE_GG)   {
        r.setRect(48, 24, 160, 144);
    }
    else    {
        r.setRect(0, 0, 256, smsvdp.lines);
    }

    painter.fillRect(rect(), Qt::transparent);
    painter.drawImage(rect(), img, r);
}

void EmuWidget::mousePressEvent(QMouseEvent *event) {
    int x, y;

    /* Scale the position to [0.0, 1.0], and multiply by the SG-1000's
    resolution. */
    x = (int)(((float)event->x() / width()) * 256);
    y = (int)(((float)event->y() / height()) * 192);

    terebi_update(x, y, 1);
    event->accept();
}

void EmuWidget::mouseReleaseEvent(QMouseEvent *event)   {
    terebi_update(0, 0, 0);
    event->accept();
}

void EmuWidget::mouseMoveEvent(QMouseEvent *event)  {
    mousePressEvent(event);
}
