/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QFileDialog>
#include <QSettings>

#include "UIPreferences.h"
#include "UIKeyInputDialog.h"
#include "CrabEmu.h"

UIPreferences::UIPreferences(QWidget *parent) : QDialog(parent) {
    QSettings settings;

    setupUi(this);

    bm = settings.value("SMS Buttonmap").toMap();
    editSMSBios->setText(settings.value("SMS BIOS ROM").toString());
    editGGBios->setText(settings.value("Game Gear BIOS ROM").toString());
    smsBiosEnabled->setChecked(settings.value("SMS BIOS Enabled").toBool());
    ggBiosEnabled->setChecked(settings.value("GG BIOS Enabled").toBool());
}

void UIPreferences::updatePreferences() {
    QSettings settings;

    settings.setValue("SMS Buttonmap", bm);
    settings.setValue("SMS BIOS ROM", editSMSBios->text());
    settings.setValue("Game Gear BIOS ROM", editGGBios->text());
    settings.setValue("SMS BIOS Enabled", smsBiosEnabled->isChecked());
    settings.setValue("GG BIOS Enabled", ggBiosEnabled->isChecked());
}

void UIPreferences::on_buttonSMS1Up_clicked()   {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Up", key);
}

void UIPreferences::on_buttonSMS1Down_clicked() {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Down", key);
}

void UIPreferences::on_buttonSMS1Left_clicked() {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Left", key);
}

void UIPreferences::on_buttonSMS1Right_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Right", key);
}

void UIPreferences::on_buttonSMS1A_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Button 1", key);
}

void UIPreferences::on_buttonSMS1B_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "SMS Button 2", key);
}

void UIPreferences::on_buttonSMS2Up_clicked()   {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Up", key);
}

void UIPreferences::on_buttonSMS2Down_clicked() {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Down", key);
}

void UIPreferences::on_buttonSMS2Left_clicked() {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Left", key);
}

void UIPreferences::on_buttonSMS2Right_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Right", key);
}

void UIPreferences::on_buttonSMS2A_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Button 1", key);
}

void UIPreferences::on_buttonSMS2B_clicked()    {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 2", "SMS Button 2", key);
}

void UIPreferences::on_buttonGGStart_clicked()  {
    int key = getKeyInput();

    if(!key)
        return;

    setKey("Qt Keyboard 1", "Game Gear Start", key);
}

void UIPreferences::on_buttonBrowseSMSBios_clicked()    {
    QString fn;

    fn = QFileDialog::getOpenFileName(QApplication::activeWindow(),
                                      tr("Select a SMS BIOS Rom"), "",
                                      tr("SMS Roms (*.sms)"), 0,
                                      QFileDialog::DontUseSheet);

    /* If a file was not selected, then don't try to open one. */
    if(fn.isEmpty())    {
        return;
    }

    editSMSBios->setText(fn);
}

void UIPreferences::on_buttonBrowseGGBios_clicked() {
    QString fn;

    fn = QFileDialog::getOpenFileName(QApplication::activeWindow(),
                                      tr("Select a Game Gear BIOS Rom"), "",
                                      tr("Game Gear Roms (*.gg)"), 0,
                                      QFileDialog::DontUseSheet);

    /* If a file was not selected, then don't try to open one. */
    if(fn.isEmpty())    {
        return;
    }

    editGGBios->setText(fn);
}

int UIPreferences::getKeyInput()    {
    UIKeyInputDialog d(window());

    if(d.exec())    {
        return d.getKeyPressed();
    }

    return 0;
}

void UIPreferences::setKey(const char *buttonmap, const char *button,
                           int key)     {
    QVariantMap map;
    
    map = bm.value(buttonmap).toMap();
    
    map[button] = key;
    bm[buttonmap] = map;
}
