/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008, 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QApplication>
#include <QMutex>

#include <QMessageBox>

#include "gui/UICrabEmu.h"
#include "EmuWidget.h"

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"

static bool resized = false;
extern "C" sms_vdp_t smsvdp;
extern "C" int sms_initialized;
extern "C" int sms_console;
static UICrabEmu *mainWindow;

QMutex memoryLock;

double GetAbsoluteTime(void)    {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (double)tv.tv_sec + (1.0E-6 * tv.tv_usec);
}

void render_scene(void) {
    QImage *img = mainWindow->emuwidget->retainImage();
    uchar *bits = img->bits();

    /* Simple, but effective. */
    memcpy(bits, smsvdp.framebuffer, 256 * 256 * 4);

    mainWindow->emuwidget->releaseImage();
}

int emu_frame(int run)  {
    memoryLock.lock();
    run = sms_frame(run);
    memoryLock.unlock();

    return run;
}

void gui_set_aspect(float x __UNUSED__, float y __UNUSED__) {
    /* XXXX: Do something about this function sometime. */
    resized = true;
}

void gui_set_title(const char *str) {
    mainWindow->setWindowTitle(QString::fromUtf8(str));
}

int main(int argc, char *argv[])    {
    QApplication app(argc, argv);

    QApplication::setOrganizationName("ljsdcdev");
    QApplication::setOrganizationDomain("ljsdcdev.net");
    QApplication::setApplicationName("CrabEmu");

    mainWindow = new UICrabEmu;
    mainWindow->setWindowTitle(app.applicationName());
    mainWindow->show();

    QObject::connect(&app, SIGNAL(lastWindowClosed()), &app, SLOT(quit()));

    return app.exec();
}
